from cqi import CQiClient
from cqi.errors import CQiException
from cqi.status import CQiStatus
from flask import session
from inspect import signature
from threading import Lock
from typing import Callable, Dict, List
from app import socketio
from app.decorators import socketio_login_required
from . import NAMESPACE as ns
from .extensions import CQI_EXTENSION_FUNCTION_NAMES
from . import extensions as extensions_module


CQI_FUNCTION_NAMES: List[str] = [
    'ask_feature_cl_2_3',
    'ask_feature_cqi_1_0',
    'ask_feature_cqp_2_3',
    'cl_alg2cpos',
    'cl_attribute_size',
    'cl_cpos2alg',
    'cl_cpos2id',
    'cl_cpos2lbound',
    'cl_cpos2rbound',
    'cl_cpos2str',
    'cl_cpos2struc',
    'cl_drop_attribute',
    'cl_id2cpos',
    'cl_id2freq',
    'cl_id2str',
    'cl_idlist2cpos',
    'cl_lexicon_size',
    'cl_regex2id',
    'cl_str2id',
    'cl_struc2cpos',
    'cl_struc2str',
    'corpus_alignment_attributes',
    'corpus_charset',
    'corpus_drop_corpus',
    'corpus_full_name',
    'corpus_info',
    'corpus_list_corpora',
    'corpus_positional_attributes',
    'corpus_properties',
    'corpus_structural_attribute_has_values',
    'corpus_structural_attributes',
    'cqp_drop_subcorpus',
    'cqp_dump_subcorpus',
    'cqp_fdist_1',
    'cqp_fdist_2',
    'cqp_list_subcorpora',
    'cqp_query',
    'cqp_subcorpus_has_field',
    'cqp_subcorpus_size',
    'ctrl_bye',
    'ctrl_connect',
    'ctrl_last_general_error',
    'ctrl_ping',
    'ctrl_user_abort'
]


@socketio.on('cqi', namespace=ns)
@socketio_login_required
def cqi_over_sio(fn_name: str, fn_args: Dict = {}):
    try:
        cqi_client: CQiClient = session['cqi_over_sio']['cqi_client']
        cqi_client_lock: Lock = session['cqi_over_sio']['cqi_client_lock']
    except KeyError:
        return {'code': 424, 'msg': 'Failed Dependency'}
    if fn_name in CQI_FUNCTION_NAMES:
        fn: Callable = getattr(cqi_client.api, fn_name)
    elif fn_name in CQI_EXTENSION_FUNCTION_NAMES:
        fn: Callable = getattr(extensions_module, fn_name)
    else:
        return {'code': 400, 'msg': 'Bad Request'}
    for param in signature(fn).parameters.values():
        if param.default is param.empty:
            if param.name not in fn_args:
                return {'code': 400, 'msg': 'Bad Request'}
        else:
            if param.name not in fn_args:
                continue
        if type(fn_args[param.name]) is not param.annotation:
            return {'code': 400, 'msg': 'Bad Request'}
    cqi_client_lock.acquire()
    try:
        fn_return_value = fn(**fn_args)
    except BrokenPipeError as e:
        return {'code': 500, 'msg': 'Internal Server Error'}
    except CQiException as e:
        return {
            'code': 502,
            'msg': 'Bad Gateway',
            'payload': {
                'code': e.code,
                'desc': e.description,
                'msg': e.__class__.__name__
            }
        }
    finally:
        cqi_client_lock.release()
    if isinstance(fn_return_value, CQiStatus):
        payload = {
            'code': fn_return_value.code,
            'msg': fn_return_value.__class__.__name__
        }
    else:
        payload = fn_return_value
    return {'code': 200, 'msg': 'OK', 'payload': payload}
